/*
 * Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * * Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 *
 * * Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in
 *   the documentation and/or other materials provided with the
 *   distribution.  
 *
 * * Neither the name of the University of California, Berkeley nor
 *   the names of its contributors may be used to endorse or promote
 *   products derived from this software without specific prior 
 *   written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package blog;

import java.util.*;
import java.io.PrintStream;


/** 
 * Implementation of NonRandomFunction for 0-ary nonrandom functions, i.e., 
 * constant symbols pointing to guaranteed objects.
 */
public class NonRandomConstant extends NonRandomFunction {
    /**
     * Creates a new NonRandomConstant whose value will be specified later.
     *
     * @param immediatelyDefined  true if this is a kind of constant symbol 
     *                            that is always defined as soon as it is 
     *                            declared.  This is the case for built-in 
     *                            symbols and symbols introduced in guaranteed 
     *                            object statements.   
     */
    public NonRandomConstant(String fname, Type ret_type, 
                             boolean immediatelyDefined) {
	super( fname, Collections.EMPTY_LIST, ret_type );
	this.immediatelyDefined = immediatelyDefined;
    }

    /**
     * @param value               the object referred to by this constant 
     *                            symbol
     *
     * @param immediatelyDefined  true if this is a kind of constant symbol 
     *                            that is always defined as soon as it is 
     *                            declared.  This is the case for built-in 
     *                            symbols and symbols introduced in guaranteed 
     *                            object statements.   
     */
    public NonRandomConstant(String fname, Type ret_type, Object value, 
                             boolean immediatelyDefined) {

	super( fname, Collections.EMPTY_LIST, ret_type );
	this.value = value;
	this.immediatelyDefined = immediatelyDefined;
    }

    /**
     * Sets the value that this non-random constant denotes.
     */
    public void setValue(Object v) {
	value = v;
    }

    /** 
     * Returns the object referred to by this constant symbol, or null if 
     * the value of this non-random constant has not been set yet.  
     *
     * @param a         An empty list, since this is a 0-ary function.
     */
    public Object getValue( List a ){
	return value;
    }

    /** 
     * Returns the object referred to by this constant symbol, or null if 
     * the value of this non-random constant has not been set yet.
     */
    public Object getValue(){
	return value;
    }

    /**
     * Returns true if this constant symbol is of a kind that is always 
     * defined as soon as it is declared.  This is true if this is a built-in 
     * constant symbol or a symbol introduced in a guaranteed object statement.
     */
    public boolean isImmediatelyDefined() {
	return immediatelyDefined;
    }

    /**
     * Prints a description of this non-random constant to the given stream.
     */
    public void print(PrintStream s) {
	s.print("nonrandom ");
	s.print(getRetType());
	s.print(" ");
	s.print(getName());
	s.print(" -> ");
	s.println(value);
    }
    
    private Object value;
    private boolean immediatelyDefined;
}

