/*
 * Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * * Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 *
 * * Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in
 *   the documentation and/or other materials provided with the
 *   distribution.  
 *
 * * Neither the name of the University of California, Berkeley nor
 *   the names of its contributors may be used to endorse or promote
 *   products derived from this software without specific prior 
 *   written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package blog;

import java.util.*;
import java.io.PrintStream;

/** Represents random functions, whose value for the given tuple of
 * arguments changes from world to world.
 *
 * @see blog.Function
 */
public class RandomFunction extends Function {


    public RandomFunction
	( String fname, 
	  List arg_types, 
	  Type ret_type, 
	  DependencyModel depmodel) {

	super( fname, arg_types, ret_type );
	this.depmodel = depmodel;

    }

    /**
     * Returns the variables that stand for the function arguments in this 
     * function's dependency model.
     *
     * @return List of String objects representing the variables.
     */
    public List getArgVars() {
	if (argVars == null) {
	    throw new IllegalStateException
		("Argument variables not set for " + this);
	}
	return argVars;
    }

    /**
     * Sets the variables that will stand for the function arguments in 
     * this function's dependency model.
     *
     * @param vars List of String objects representing the variables
     */
    public void setArgVars(List vars) {
	argVars = vars;
    }

    public DependencyModel getDepModel( ) {   

	return depmodel;

    }


    public void setDepModel( DependencyModel depmodel ){

	this.depmodel = depmodel;

    }


    public boolean hasDepModel( ){

	return ( depmodel != null );

    }

    /**
     * Prints the dependency statement for this function to the given stream.
     */
    public void printDepStatement(PrintStream s) {
	s.print(getName());
	if (!argVars.isEmpty()) {
	    s.print("(");
	    s.print(argVars.get(0));
	    for (int i = 1; i < argVars.size(); ++i) {
		s.print(", ");
		s.print(argVars.get(i));
	    }
	    s.print(")");
	}
	s.println(":");
	if (depmodel != null) {
	    depmodel.print(s);
	}
    }

    public Object getValueInWorld(List args, PartialWorld w, 
				  ValueChooser chooser, Set parents) {
	if (!w.areValidObjects(args)) {
	    throw new IllegalArgumentException
		("Some of the following arguments do not exist: " + args);
	}

	if (args.contains(Model.NULL)) {
	    return Model.NULL;
	}

	List specificArgs = Function.getSpecificArgs(args, w, chooser); 
	if (specificArgs == null) {
	    return null;
	}

	RandFuncAppVar var = new RandFuncAppVar(this, specificArgs);
	if (!w.isInstantiated(var)) {
	    if (chooser == null) {
		return null;
	    }
	    var.ensureInstAndSupported(w, chooser);
	}
	if (parents != null) {
	    parents.add(var);
	}

	return w.getValue(var);
    }

    private List argVars = null; // of String
    private DependencyModel depmodel = null;
}



